from typing import Any, AsyncGenerator, Callable, Coroutine, Generator, TypeVar, Union

T = TypeVar("T")

CallableProvider = Callable[..., T]
CoroutineProvider = Callable[..., Coroutine[Any, Any, T]]
GeneratorProvider = Callable[..., Generator[T, None, None]]
AsyncGeneratorProvider = Callable[..., AsyncGenerator[T, None]]

DependencyProviderType = Union[
    CallableProvider[T],
    CoroutineProvider[T],
    GeneratorProvider[T],
    AsyncGeneratorProvider[T],
]


DependencyProvider = Union[
    AsyncGeneratorProvider[Any],
    CoroutineProvider[Any],
    GeneratorProvider[Any],
    CallableProvider[Any],
]



from __future__ import annotations

from typing import Dict, Hashable, List, TypeVar, Union

from di.api.scopes import Scope
from di.exceptions import DuplicateScopeError, UnknownScopeError

KT = TypeVar("KT", bound=Hashable)
VT = TypeVar("VT")
T = TypeVar("T")


class ScopeMap(Dict[Scope, Dict[KT, VT]]):
    """Mapping-like structure to hold cached values.

    The important aspect of this is that we may have thousands of dependencies, but generally only a couple (<10) scopes.
    So this is designed to operate in ~ O(S) time, where S is the number of scopes.
    Imporantly, when we enter or exit a scope (which can happen very frequently, e.g. requests in a web framework),
    no iteration is required: we can simply pop all keys in that scope.
    """

    def get_key(self, key: KT, *, scope: Scope, default: T) -> Union[VT, T]:
        for current_scope, scopemap in self.items():
            if key in scopemap:
                return scopemap[key]
            if current_scope == scope:
                break
        return default

    def set(self, key: KT, value: VT, *, scope: Scope) -> None:
        try:
            self[scope][key] = value
        except KeyError:
            raise UnknownScopeError(
                f"The scope {scope} was not found. Did you forget to enter it?"
            )

    def add_scope(self, scope: Scope) -> None:
        if scope in self:
            raise DuplicateScopeError(f"The scope {scope} already exists!")
        self[scope] = {}  # type: ignore

    def __repr__(self) -> str:  # pragma: no cover, used only for debugging
        values: List[str] = []
        for scope, mapping in self.items():
            for k, v in mapping.items():
                values.append(f'{repr(k)}: {repr(v)} @ scope="{scope}"')
        return "{" + ", ".join(values) + "}"



from __future__ import annotations

from contextlib import AsyncExitStack, ExitStack
from typing import Any, Dict, Iterable, Mapping, NamedTuple, Optional, Tuple, Union

from graphlib2 import TopologicalSorter

from di._utils.scope_map import ScopeMap
from di._utils.task import ExecutionState, Task, gather_new_tasks
from di._utils.types import CacheKey
from di.api.executor import Task as ExecutorTask
from di.api.providers import DependencyProvider
from di.api.scopes import Scope
from di.api.solved import SolvedDependant

Results = Dict[int, Any]


class SolvedDependantCache(NamedTuple):
    """Private data that the Container attaches to SolvedDependant"""

    root_task: Task
    topological_sorter: TopologicalSorter[Task]


def plan_execution(
    stacks: Mapping[Scope, Union[AsyncExitStack, ExitStack]],
    cache: ScopeMap[CacheKey, Any],
    solved: SolvedDependant[Any],
    *,
    values: Optional[Mapping[DependencyProvider, Any]] = None,
) -> Tuple[Dict[int, Any], Iterable[Optional[ExecutorTask]], Any, Task,]:
    solved_dependency_cache: "SolvedDependantCache" = solved.container_cache
    ts = solved_dependency_cache.topological_sorter.copy()
    results: "Results" = {}
    execution_state = ExecutionState(
        values=values or {},
        stacks=stacks,
        results=results,
        toplogical_sorter=ts,
        cache=cache,
    )
    return (
        results,
        gather_new_tasks(execution_state),
        execution_state,
        solved_dependency_cache.root_task,
    )
